﻿using UnityEngine;
using System.Collections;

public class GUIScript : MonoBehaviour {

	public GUISkin MySkin;
	public TileMap MainMap;
	public Texture VictoryTex;

	public Transform MenuBlocker;
	public Transform MenuHelp;

	GUIStyle RedTextStyle;

	int LastSelectedLevel = 0;
	int LevelInt = 0;
	int VisibleLevelNumber = 1;

	bool IsInHelp = true;

	bool LastWasRandom = false;

	int RandomDifficulty = 0;
	string[] RandomDifficultyStrings = new string[] {"Easy", "Medium", "Hard", "Unknown"};
	
	// Use this for initialization
	void Start () {
		
		RedTextStyle = new GUIStyle ();
		RedTextStyle.normal.textColor = new Color (1.0f, 0.0f, 0.0f);
	}

	public void SetVisibleLevel(int NewLevel) {
		VisibleLevelNumber = NewLevel;
	}

	void OnGUI () {

		GUI.skin = MySkin;

		if (MainMap == null) {
			return;
		}

		if (GUI.Button (new Rect (150, 530, 120, 50), IsInHelp ? "Start" : "Help")) {
			IsInHelp = !IsInHelp;
		}

		MainMap.IsInHelp = IsInHelp;

		if (IsInHelp) {

			MenuBlocker.gameObject.SetActive(true);
			MenuHelp.gameObject.SetActive(true);
			return;
		} else {

			MenuBlocker.gameObject.SetActive(false);
			MenuHelp.gameObject.SetActive(false);
		}

		bool DebugButtons = false;

		int RunX = 130;
		int SpaceY = 30;
		int RunY = 10;
		int NumElement = 5;

		GUI.Box(new Rect(RunX,RunY,120,NumElement*SpaceY), "Current Path");

		RunY += SpaceY;

		if(GUI.Button(new Rect(RunX+10,RunY,100,20), "Reset")) {
			MainMap.ResetAll();
		}

		RunY += SpaceY;

		if(GUI.Button(new Rect(RunX+10,RunY,100,20), "Undo")) {
			MainMap.Undo();
		}
		
		RunY += SpaceY * 2;

		/*if(GUI.Button(new Rect(RunX+10,RunY,100,20), "Scramble")) {
			MainMap.WantScramble();
		}

		RunY += SpaceY;*/

		if (MainMap.IsWinState()) {

			GUI.DrawTexture(new Rect(RunX, RunY, 120, 120), VictoryTex, ScaleMode.ScaleToFit, true);

			RunY += 80;

			if(!LastWasRandom) {

				if(LevelInt < 15) {

					if (GUI.Button (new Rect (RunX+10, RunY, 100, 20), "Next Level")) {
						if(LevelInt == (VisibleLevelNumber-1) && VisibleLevelNumber < 16) {
							++VisibleLevelNumber;
						}
						++LevelInt;
					}
				}
				
				RunY += SpaceY;
			}
		}

		RunY = 10;

		if (LastWasRandom) {
			GUI.Label (new Rect (20, RunY, 100, 20), "Levels");
		} else {
			GUI.Label (new Rect (20, RunY, 100, 20), "Levels", RedTextStyle);
		}
		
		RunY += SpaceY;

		string[] LevelStrings = new string[VisibleLevelNumber];
		for(int i=0; i<VisibleLevelNumber; ++i) {
			LevelStrings[i] = "Level " + (i+1);
		}

		LevelInt = GUI.SelectionGrid(new Rect(25, RunY, 100, VisibleLevelNumber * SpaceY), LevelInt, LevelStrings, 1);

		if (LevelInt != LastSelectedLevel) {
			LastSelectedLevel = LevelInt;
			if(LevelInt >=0 && LevelInt<VisibleLevelNumber) {
				MainMap.StartPremadeLevel(LevelInt + 1, true);
			}
			LastWasRandom = false;
		}

		RunY = 300;

		if (LastWasRandom) {
			GUI.Label (new Rect (RunX + 10, RunY, 100, 20), "Random", RedTextStyle);
		} else {
			GUI.Label (new Rect (RunX + 10, RunY, 100, 20), "Random");
		}
		
		RunY += SpaceY;

		if(GUI.Button(new Rect(RunX + 15,RunY,100,40), "New Random")) {
			MainMap.StartRandomLevel(RandomDifficulty);
			LastWasRandom = true;
			LevelInt = -1;
			LastSelectedLevel = -1;
		}

		RunY += SpaceY + 20;

		int SelBoxLength = RandomDifficultyStrings.Length * SpaceY;

		RandomDifficulty = GUI.SelectionGrid(new Rect(RunX + 25, RunY, 80, SelBoxLength), RandomDifficulty, RandomDifficultyStrings, 1);

		RunY += SelBoxLength + SpaceY;

		RunY = 10;
		RunX = 800;
		
		if (DebugButtons) {
			if (GUI.Button (new Rect (RunX+10, RunY, 100, 20), "Solution")) {
				MainMap.ViewSolution ();
			}
			
			RunY += SpaceY;
			
			if (GUI.Button (new Rect (RunX+10, RunY, 100, 20), "Write Log")) {
				MainMap.WriteCellsToLog ();
			}
			
			RunY += SpaceY;
			
			if(LevelInt < 15) {
				
				if (GUI.Button (new Rect (RunX+10, RunY, 100, 20), "Force Next")) {
					if(LevelInt == (VisibleLevelNumber-1) && VisibleLevelNumber < 16) {
						++VisibleLevelNumber;
					}
					++LevelInt;
				}
			}
			
			RunY += SpaceY;
		}
	}
}
