﻿using UnityEngine;
using System.Collections;
using System.Collections.Generic;

public enum MoveDir {
	None,
	Left,
	Right,
	LTop,
	RTop,
	LDown,
	RDown
}

public class GraphConnect {
	public bool[] G = new bool[7];

	public GraphConnect() {
		for(int i=0; i<7; ++i) {
			this.G[i] = false;
		}
	}

	public GraphConnect(bool L, bool R, bool LT, bool RT, bool LD, bool RD) {
		this.G[0] = false;
		this.G[(int)MoveDir.Left] = L;
		this.G[(int)MoveDir.Right] = R;
		this.G[(int)MoveDir.LTop] = LT;
		this.G[(int)MoveDir.RTop] = RT;
		this.G[(int)MoveDir.LDown] = LD;
		this.G[(int)MoveDir.RDown] = RD;
	}

	public GraphConnect(int L, int R, int LT, int RT, int LD, int RD) {
		this.G[0] = false;
		this.G[(int)MoveDir.Left] = (L != 0);
		this.G[(int)MoveDir.Right] = (R != 0);
		this.G[(int)MoveDir.LTop] = (LT != 0);
		this.G[(int)MoveDir.RTop] = (RT != 0);
		this.G[(int)MoveDir.LDown] = (LD != 0);
		this.G[(int)MoveDir.RDown] = (RD != 0);
	}

	public void SetConnect(MoveDir Dir, bool Value) {
		this.G[(int)Dir] = Value;
	}

	public bool IsConnected(MoveDir Dir) {
		return this.G [(int)Dir];
	}
}

public class SaveState {
	public List< List<TileScript> > Cells;

	public SaveState( List< List<TileScript> > InCells) {

		Cells = new List< List<TileScript> > ();

		for (int x = 0; x<InCells.Count; ++x) {

			Cells.Add ( new List<TileScript>() );

			for (int y = 0; y<InCells[x].Count; ++y) {				
				Cells[x].Add(InCells[x][y]);
			}
		}
	}

	public void RestoreState( List< List<TileScript> > InCells) {

		for (int x = 0; x<InCells.Count; ++x) {

			for (int y = 0; y<InCells[x].Count; ++y) {				
				InCells[x][y] = Cells[x][y];
			}
		}
	}
}

public class TileMap : MonoBehaviour {

	public Texture2D TileTex;
	public Texture2D GraphTex;
	public Texture2D RedOutTex;

	public Transform BlackQuad;
	public GUIScript Interface;

	public Transform SoundPuzzleStart;
	public Transform SoundPuzzleWin;

	Sprite MainSprite;
	Sprite WarnSprite;
	
	List<TileScript> Sprites;
	List< List<TileScript> > Cells;

	List<TileScript> SpritesToMove;

	int CellNumX = 8;
	int CellNumY = 8;

	float ResetTime = 0.0f;
	int ResetNum = 0;

	bool NeedUpdateEnergy = true;
	bool IsInit = false;
	bool HaveWon = false;

	public bool IsInHelp = false;

	public bool IsWinState() {
		return HaveWon;
	}

	float GridSize;
	float GridSizeY;

	Color BackColor;
	public float TimeBlackScreen;

	TileScript PrevOverideTile = null;
	TileScript DragTile = null;
	bool WasMouseClic = false;
	Vector2 LastNearestPos = new Vector2 ();
	Vector2 DragLastDir = new Vector2 ();

	Vector2 TwistDir;

	List< List<Sprite> > GraphSprites;

	SaveState SolutionSave;
	SaveState CurrentSave;

	List<SaveState> History;

	Levels MainLevels;

	// Use this for initialization
	void Start () {

		GridSize = 0.9f;
		GridSizeY = GridSize * Mathf.Sqrt (3.0f) / 2.0f;

		TimeBlackScreen = 0.0f;

		MainLevels = new Levels(this);

		History = new List<SaveState> ();

		Sprites = new List<TileScript>();
		Cells = new List< List<TileScript> > ();
		SpritesToMove = new List<TileScript>();
		BackColor = new Color (0.3f, 0.3f, 0.3f);

		TwistDir = new Vector2(0.5f, Mathf.Sqrt (3.0f) / 2.0f);
		TwistDir.Normalize();
	
		MainSprite = Sprite.Create (TileTex, new Rect (0,0,64,64), new Vector2 (0.5f,0.5f), 64);
		WarnSprite = Sprite.Create (RedOutTex, new Rect (0,0,64,64), new Vector2 (0.5f,0.5f), 64);

		GraphSprites = new List< List<Sprite> > ();
		for (int x = 0; x<8; ++x) {
			
			GraphSprites.Add ( new List<Sprite>() );
			
			for (int y = 0; y<4; ++y) {

				Rect GraphRect = new Rect (x * 64, y * 64, 64, 64);

				Sprite GraphSprite = Sprite.Create (GraphTex, GraphRect, new Vector2 (0.5f,0.5f), 64);

				GraphSprites[x].Add( GraphSprite );

			}
		}


		for (int x = 0; x<CellNumX; ++x) {

			Cells.Add ( new List<TileScript>() );

			for (int y = 0; y<CellNumY; ++y) {
					
				Cells[x].Add( null );

				AddBackSprite(x,y);

			}
		}

		StartCreateLevel ();

		IsInit = true;

	}

	void ClearLevel() {

		HaveWon = false;
		NeedUpdateEnergy = true;
		History.Clear ();

		for (int i = 0; i<Sprites.Count; ++i) {			
			Destroy(Sprites[i].TileObj);
		}

		for (int x = 0; x<CellNumX; ++x) {
			for (int y = 0; y<CellNumY; ++y) {				
				Cells[x][y] = null;
			}
		}

		Sprites.Clear();
	}

	public void SetMainStart(int x, int y) {
		TileScript StarTile = Cells[x][y];
		if (StarTile != null) {
			StarTile.IsStar = true;
		}
	}

	public void StartPremadeLevel(int Index, bool PlaySound) {

		if (PlaySound) {
			SoundPuzzleStart.audio.Play ();
		}

		ClearLevel ();

		MainLevels.LoadLevel (Index);

		BuildGraphSprites ();

		SolutionSave = new SaveState (Cells);
		CurrentSave = new SaveState (Cells);
	}

	public void StartRandomLevel(int Index) {

		SoundPuzzleStart.audio.Play ();

		ClearLevel ();

		//Random.seed = 124 + ResetNum * 1142;
		//Random.seed = 50641021 - 4512;

		int TileNum = Random.Range(5, 9);
		float ConRate = 0.5f;
		int RandomLoop = 500;

		if (Index == 1) {
			TileNum = Random.Range(10, 14);
			ConRate = 0.5f;
			RandomLoop = 200;
		} else if (Index == 2) {
			TileNum = Random.Range(15, 19);
			ConRate = 0.7f;
			RandomLoop = 1000;
		} else if (Index == 3) {
			TileNum = Random.Range(20, 25);
			ConRate = 1.0f;
			RandomLoop = 4000;
		}
		
		int MainStarX = 4;
		int MainStarY = 4;
		AutoAddSprites2 (MainStarX, MainStarY, TileNum);
		
		SetMainStart(MainStarX, MainStarY);
		
		CrankConnections (ConRate);

		BuildGraphSprites ();

		SolutionSave = new SaveState (Cells);
		
		ScrambleTest ( RandomLoop );

		CurrentSave = new SaveState (Cells);
	}

	public void StartCreateLevel() {

		StartPremadeLevel (1, false);
		//StartRandomLevel (0);
	}

	public void WriteCellsToLog() {

		int LastStartFoundX = 0;
		int LastStartFoundY = 0;

		string FullDesc = "";
		for (int i = 0; i<Sprites.Count; ++i) {		
			TileScript CurTile = Sprites[i];

			string SpriteDesc = "";
			SpriteDesc += "AddSprite (";
			SpriteDesc += CurTile.PosX;
			SpriteDesc += ",";
			SpriteDesc += CurTile.PosY;
			SpriteDesc += ", ";
			for(int c=1; c<7; ++c) {
				MoveDir Dir = (MoveDir)c;
				bool Connect = CurTile.Connections.IsConnected(Dir);
				SpriteDesc += Connect ? "1" : "0";
				if(c != 6) SpriteDesc += ",";
			}
			SpriteDesc += ");";
			SpriteDesc += "\n";

			//AddSprite (3,1, 1,0,0,0,0,0);

			FullDesc += SpriteDesc;

			if(CurTile.IsStar) {
				LastStartFoundX = CurTile.PosX;
				LastStartFoundY = CurTile.PosY;
			}
		}

		FullDesc += "SetMainStart(" + LastStartFoundX + "," + LastStartFoundY + ");";
		FullDesc += "\n";

		Debug.Log(FullDesc);
	}

	void UpdateAfterRestoreState() {

		for (int x = 0; x<CellNumX; ++x) {
			for (int y = 0; y<CellNumY; ++y) {
				
				TileScript CurTile = Cells [x] [y];
				if (CurTile != null) {
					CurTile.PosX = x;
					CurTile.PosY = y;
				}
			}
		}
		
		UpdateTileBasicPos ();
		
		HaveWon = false;
		NeedUpdateEnergy = true;
	}

	public void Undo() {
		if (!IsInit || History.Count == 0) {
			return;
		}
		int TopIndex = History.Count - 1;
		SaveState TopState = History [TopIndex];
		History.RemoveAt (TopIndex);

		TopState.RestoreState (Cells);
		UpdateAfterRestoreState ();
	}

	public void ResetAll() {
		if (!IsInit) {
			return;
		}
		History.Clear ();
		CurrentSave.RestoreState (Cells);
		UpdateAfterRestoreState ();
	}

	public void ViewSolution() {
		if (!IsInit) {
			return;
		}
		History.Clear ();
		SolutionSave.RestoreState (Cells);
		UpdateAfterRestoreState ();
	}

	public MoveDir InverseDir(MoveDir Dir) {
		if(Dir == MoveDir.Left) {
			return MoveDir.Right;
		} else if(Dir == MoveDir.Right) {
			return MoveDir.Left;
		} else if(Dir == MoveDir.LTop) {
			return MoveDir.RDown;
		} else if(Dir == MoveDir.RTop) {
			return MoveDir.LDown;
		} else if(Dir == MoveDir.LDown) {
			return MoveDir.RTop;
		} else if(Dir == MoveDir.RDown) {
			return MoveDir.LTop;
		}
		return MoveDir.None;
	}

	class TileConstruct {
		
		public int PosX;
		public int PosY;
		public MoveDir AddDir;

		public TileConstruct(int InPosX, int InPosY, MoveDir InAddDir) {
			PosX = InPosX;
			PosY = InPosY;
			AddDir = InAddDir;
		}
	}

	void AutoAddSprites2( int StartX, int StartY, int Loop) {
		
		int CurX = StartX;
		int CurY = StartY;

		int AddedSprite = 0;

		List<TileConstruct> ToCheck = new List<TileConstruct>();
		ToCheck.Add (new TileConstruct(CurX, CurY, MoveDir.None));
		
		while(ToCheck.Count != 0) {

			int RandPos = Random.Range(0,ToCheck.Count - 1);
			TileConstruct CurConst = ToCheck[RandPos];
			ToCheck.RemoveAt(RandPos);

			TileScript CurTile = AddSprite (CurConst.PosX,CurConst.PosY, new GraphConnect ());
			++AddedSprite;

			if(CurTile != null && CurConst.AddDir != MoveDir.None) {
				MoveDir InvDir = InverseDir(CurConst.AddDir);
				TileScript Neight = GetNeight (CurConst.PosX,CurConst.PosY,InvDir);
				if(Neight != null) {
					CurTile.Connections.SetConnect(InvDir, true);
					Neight.Connections.SetConnect(CurConst.AddDir, true);
				}
			}

			for(int i=1; i<7; ++i) {
				MoveDir Dir = (MoveDir)i;
				int NewX = 0;
				int NewY = 0;
				if(GetMoveNewPos(CurConst.PosX,CurConst.PosY, Dir, ref NewX, ref NewY)) {
					if(Cells[NewX][NewY] == null) {

						bool AlreadyInCheck = false;
						for(int c=0; c<ToCheck.Count; ++c) {
							if(ToCheck[c].PosX == NewX && ToCheck[c].PosY == NewY) {
								AlreadyInCheck = true;
								break;
							}
						}
						if(!AlreadyInCheck) {
							ToCheck.Add(new TileConstruct(NewX, NewY, Dir));
						}
					}
				}
			}
		
			if(AddedSprite >= Loop-1) {
				break;
			}
		}
	}

	void AutoAddSprites( int StartX, int StartY, int Loop) {

		int CurX = StartX;
		int CurY = StartY;

		TileScript CurTile = AddSprite (CurX,CurY, new GraphConnect ());
		if (CurTile == null) {
			return;
		}

		int AddedSprite = 0;
		int MaxLoop = 100;

		for (int i=0; i<MaxLoop; ++i) {

			MoveDir RandDir = (MoveDir)Random.Range(1,6);
			int NewX = CurX;
			int NewY = CurY;
			TileScript Neight = GetNeight (CurX, CurY, RandDir, ref NewX, ref NewY);
			if(Neight == null && !IsOutside(NewX, NewY)) {

				TileScript NewScript = AddSprite (NewX,NewY, new GraphConnect());
				if(NewScript != null) {

					NewScript.Connections.SetConnect( InverseDir(RandDir), true);
					CurTile.Connections.SetConnect( RandDir, true);

					CurTile = NewScript;
					CurX = NewX;
					CurY = NewY;
					++AddedSprite;
				}
			}

			if(AddedSprite >= Loop-1) {
				break;
			}
		}
	}

	void CrankConnections(float Percentage) {

		for (int i=0; i<Sprites.Count; ++i) {

			TileScript CurTile = Sprites[i];
			int CurX = CurTile.PosX;
			int CurY = CurTile.PosY;
			for(int c=1; c<7; ++c) {

				int NewX = CurX;
				int NewY = CurY;
				MoveDir CurDir = (MoveDir)(c);
				TileScript Neight = GetNeight (CurX, CurY, CurDir, ref NewX, ref NewY);
				if(Neight != null && CurX <= NewX && CurY <= NewY ) {

					if(Random.Range(0.0f, 1.0f) < Percentage) {
						Neight.Connections.SetConnect( InverseDir(CurDir), true);
						CurTile.Connections.SetConnect( CurDir, true);
					}
				}
			}
		}
	}

	Vector2 GetMoveDir(Vector3 StartPos, Vector3 GoalPos) {

		Vector3 Direction = GoalPos - StartPos;
		Vector2 DragDir = new Vector2(Direction.x, Direction.y);
		return DragDir;
	}

	public void WantScramble() {
		ScrambleTest(500);
		CurrentSave = new SaveState (Cells);
		HaveWon = false;
		NeedUpdateEnergy = true;
	}

	void ScrambleTest(int Loops) {

		History.Clear ();

		List<TileScript> Starts = new List<TileScript>();
		for (int i=0; i<Sprites.Count; ++i) {
			
			TileScript CurTile = Sprites[i];
			if(CurTile.IsStar) {
				Starts.Add(CurTile);
				CurTile.IsEnergy = true;
				for(int c=1; c<7; ++c) {
					CurTile.Connections.SetConnect((MoveDir)c, true);
				}
			}
		}

		float GridFactor = Mathf.Sqrt(CellNumX * CellNumX);

		Vector3 CenterPos = CalcSpritePosition(CellNumX/2, CellNumY/2);

		int MaxLoop = Loops;
		for (int i=0; i<MaxLoop; ++i) {
			
			int RandStart = Random.Range(0,Starts.Count-1);
			TileScript Start = Starts[RandStart];
			Vector3 StartPos = CalcSpritePosition(Start.PosX, Start.PosY);

			Vector2 MoveCenter = GetMoveDir(StartPos, CenterPos);
			if(Random.Range(0.0f, 1.0f) < (1.7f * MoveCenter.magnitude / GridFactor)) {
				MoveCenter.Normalize();

				RequestMoveTile (Start, MoveCenter, false);
			} else if(Random.Range(0.0f, 1.0f) > 0.7f) {

				int RandGoal = Random.Range(0, Sprites.Count-1);
				TileScript Goal = Sprites[RandGoal];

				if(Start != Goal) {

					Vector3 GoalPos = CalcSpritePosition(Goal.PosX, Goal.PosY);
					Vector2 MoveDirNorm = GetMoveDir(StartPos, GoalPos);
					MoveDirNorm.Normalize();

					if(Random.Range(0.0f, 1.0f) > 0.7f) {
						MoveDirNorm = -MoveDirNorm;
					}

					RequestMoveTile (Start, MoveDirNorm, false);
				}
			} else {

				MoveDir RandDir = (MoveDir)Random.Range(1,6);
				MoveTile (Start, RandDir, false);
			}

			if(NeedUpdateEnergy) {
				UpdateEnergySprites ();
			}
		}
	}

	Sprite FindGraph(GraphConnect GC, int x, int y, ref bool MirrorX, ref bool MirrorY) {
		int GraphX = 2;
		int GraphY = 0;
			
		int NumConnect = 0;
		for (int i=0; i<7; ++i) {
			NumConnect += GC.G[i] == true ? 1 : 0;
		}

		bool L = GC.G [(int)MoveDir.Left];
		bool R = GC.G [(int)MoveDir.Right];
		bool LT = GC.G [(int)MoveDir.LTop];
		bool RT = GC.G [(int)MoveDir.RTop];
		bool LD = GC.G [(int)MoveDir.LDown];
		bool RD = GC.G [(int)MoveDir.RDown];
		
		if (NumConnect == 0) {
			
		} else if(NumConnect == 1) {
			GraphY = 3;
			if( L || R ) {
				GraphX = 0;
				if(R) MirrorX = true;
			} else {
				GraphX = 1;
				if(RT || RD) MirrorX = true;
				if(LD || RD) MirrorY = true;
			}
		} else if(NumConnect == 2) {
			GraphY = 3;
			if( L || R ) {
				if(L && (LT || LD) || R && (RT || RD)) {
					GraphX = 2;
				} else if(L && R) {
					GraphX = 4;
				} else {
					GraphX = 3;
				}
				if(LD || RD) MirrorY = true;
				if(R) MirrorX = true;
			} else {
				if(LT && RT || LD && RD) {
					GraphX = 5;
				} else if(LT && LD || RT && RD){
					GraphX = 6;
				} else {
					GraphX = 7;
				}
				
				if(RT) MirrorX = true;
				if(RD && !LT) MirrorY = true;
			}
		} else if(NumConnect == 3) {
			GraphY = 2;
			if( L ) {
				if( R ) {
					GraphX = 1;
					MirrorX = RT || RD;
					MirrorY = LD || RD;
				} else if( LT && RT || LD && RD) {
					GraphX = 0;
					MirrorY = LD;
				} else if( LT && RD || LD && RT) {
					GraphX = 2;
					MirrorY = LD;
				} else if(LT && LD) {
					GraphX = 3;
				} else if( RT && RD) {
					GraphX = 5;
				}
			} else if(R) {
				MirrorX = true;
				if( RT && LT || RD && LD) {
					GraphX = 0;
					MirrorY = RD;
				} else if( RT && LD || RD && LT) {
					GraphX = 2;
					MirrorY = RD;
				} else if( RT && RD) {
					GraphX = 3;
				} else if( LT && LD) {
					GraphX = 5;
				}
			} else {
				GraphX = 7;				
				if(!LT || !LD) MirrorX = true;
				if(!LT || !RT) MirrorY = true;
			}
		} else if(NumConnect == 4) {
			GraphY = 1;
			if( !L || !R ) {
				if(!L && (!LT || !LD) || !R && (!RT || !RD)) {
					GraphX = 2;
				} else if(!L && !R) {
					GraphX = 4;
				} else {
					GraphX = 3;
				}
				if(!LD || !RD) MirrorY = true;
				if(!R) MirrorX = true;
			} else {
				if(!LT && !RT || !LD && !RD) {
					GraphX = 5;
				} else if(!LT && !LD || !RT && !RD){
					GraphX = 6;
				} else {
					GraphX = 7;
				}
				
				if(!RT) MirrorX = true;
				if(!RD && LT) MirrorY = true;
			}
		} else if(NumConnect == 5) {
			GraphY = 1;
			if( !L || !R ) {
				GraphX = 0;
				if(!R) MirrorX = true;
			} else {
				GraphX = 1;
				if(!RT || !RD) MirrorX = true;
				if(!LD || !RD) MirrorY = true;
			}
		} else if(NumConnect == 6) {
			GraphX = 0;
			GraphY = 0;
		}
		
		//Rect GraphRect = new Rect (GraphX * 64, GraphY * 64, 64, 64);
		//Sprite GraphSprite = Sprite.Create (GraphTex, GraphRect, new Vector2 (0.5f,0.5f), 64);
		
		return GraphSprites[GraphX][GraphY];
	}

	GameObject CreateGraphSprite(GraphConnect GC, int x, int y) {

		bool MirrorX = false;
		bool MirrorY = false;
		Sprite GraphSprite = FindGraph (GC, x, y, ref MirrorX, ref MirrorY);

		GameObject CurGraphObj = new GameObject("TileGraph_"+x+"_"+y);
		SpriteRenderer CurGraphSpriteRen = CurGraphObj.AddComponent("SpriteRenderer") as SpriteRenderer;
		CurGraphSpriteRen.sprite = GraphSprite;
		CurGraphObj.transform.localScale = new Vector3 (MirrorX ? -1.0f : 1.0f, MirrorY ? -1.0f : 1.0f, 1.0f);

		return CurGraphObj;
	}

	void UpdateGraphSprite(TileScript Tile) {

		bool MirrorX = false;
		bool MirrorY = false;
		Sprite GraphSprite = FindGraph (Tile.Connections, Tile.PosX, Tile.PosY, ref MirrorX, ref MirrorY);

		Tile.TileGraphObj.transform.localScale = new Vector3 (MirrorX ? -1.0f : 1.0f, MirrorY ? -1.0f : 1.0f, 1.0f);

		SpriteRenderer SR = Tile.TileGraphObj.renderer as SpriteRenderer;
		SR.sprite = GraphSprite;
	}

	public GameObject CreateStarPicture(TileScript Tile) {
		Rect GraphRect = new Rect (1 * 64, 0 * 64, 64, 64);
		
		Sprite GraphSprite = Sprite.Create (GraphTex, GraphRect, new Vector2 (0.5f,0.5f), 64);
		GameObject CurGraphObj = new GameObject("TileStar_"+Tile.PosX+"_"+Tile.PosY);
		SpriteRenderer CurGraphSpriteRen = CurGraphObj.AddComponent("SpriteRenderer") as SpriteRenderer;
		CurGraphSpriteRen.sprite = GraphSprite;
		CurGraphSpriteRen.color = new Color (0.3f, 1.0f, 0.3f);

		return CurGraphObj;
	}

	public GameObject CreateWarnPicture(TileScript Tile, int Index) {
		GameObject CurGraphObj = new GameObject("TileWarn_"+Tile.PosX+"_"+Tile.PosY+"_"+Index);
		SpriteRenderer CurGraphSpriteRen = CurGraphObj.AddComponent("SpriteRenderer") as SpriteRenderer;
		CurGraphSpriteRen.sprite = WarnSprite;
		
		return CurGraphObj;
	}
	void BuildGraphSprites() {

		for (int x = 0; x<CellNumX; ++x) {
			for (int y = 0; y<CellNumY; ++y) {

				TileScript CurTile = Cells[x][y];
				if(CurTile != null) {
					for(int child = 0; child<CurTile.transform.childCount; ++child) {
						Transform CurChild = CurTile.transform.GetChild(child);
						Destroy(CurChild.gameObject);
					}

					GameObject CurGraphObj = CreateGraphSprite (CurTile.Connections, x, y);
					CurGraphObj.transform.parent = CurTile.TileObj.transform;
					CurGraphObj.transform.localPosition = new Vector3 (0.0f, 0.0f, -1.0f);

					CurTile.TileGraphObj = CurGraphObj.renderer as SpriteRenderer;

					if(CurTile.IsStar) {
						GameObject CurStarObj = CreateStarPicture (CurTile);
						CurStarObj.transform.parent = CurTile.TileObj.transform;
						CurStarObj.transform.localPosition = new Vector3 (0.0f, 0.0f, -2.0f);
					} else {
						for(int c=1; c<7; ++c) {
							GameObject CurWarnObj = CreateWarnPicture (CurTile, c);
							CurWarnObj.transform.parent = CurTile.TileObj.transform;
							CurWarnObj.transform.localPosition = new Vector3 (0.0f, 0.0f, -2.0f);
							CurTile.RedWarn[c] = CurWarnObj;

							Quaternion TmpRot = CurWarnObj.transform.localRotation;
							float RotAngle = 0.0f;
							MoveDir CurDir = (MoveDir)c;
							if(CurDir == MoveDir.LTop) {
								RotAngle = 60.0f;
							} else if(CurDir == MoveDir.LTop) {
								RotAngle = 60.0f;
							} else if(CurDir == MoveDir.RTop) {
								RotAngle = 120.0f;
							} else if(CurDir == MoveDir.Right) {
								RotAngle = 180.0f;
							} else if(CurDir == MoveDir.RDown) {
								RotAngle = 240.0f;
							} else if(CurDir == MoveDir.LDown) {
								RotAngle = 300.0f;
							}
							TmpRot.eulerAngles = new Vector3(0.0f, 0.0f, -RotAngle);
							CurWarnObj.transform.localRotation = TmpRot;
						}
					}
				}
			} 
		} 
	}

	void UpdateStartConnexions(TileScript Tile) {

		for (int i=1; i<7; ++i) {
			MoveDir Dir = (MoveDir)i;
			MoveDir InvDir = InverseDir(Dir);
			Tile.Connections.SetConnect(Dir, false);

			TileScript Neight = GetNeight(Tile.PosX, Tile.PosY, Dir);
			if(Neight != null) {
				if(Neight.Connections.G[(int)InvDir]) {
					Tile.Connections.SetConnect(Dir, true);
				}
			}
		}

		UpdateGraphSprite (Tile);
	}

	void UpdateEnergySprites() {

		List<TileScript> ToCheck = new List<TileScript>();

		for (int i=0; i<Sprites.Count; ++i) {
			TileScript Current = Sprites[i];
			Current.IsEnergy = false;
			if(Current.IsStar) {

				ToCheck.Add (Current);

				UpdateStartConnexions(Current);
			}
		}

		while (ToCheck.Count != 0) {
			TileScript Current = ToCheck[ToCheck.Count-1];
			ToCheck.RemoveAt(ToCheck.Count-1);

			if(!Current.IsEnergy) {
				Current.IsEnergy = true;

				GraphConnect GC = Current.Connections;

				for(int c=1; c<7; ++c) {
					MoveDir Dir = (MoveDir)c;
					if(GC.IsConnected(Dir)) {
						TileScript Neight = GetNeight (Current.PosX, Current.PosY, Dir);
						if(Neight != null && !Neight.IsEnergy) {
							GraphConnect OGC = Neight.Connections;
							if(OGC.IsConnected(InverseDir(Dir))) {
								ToCheck.Add(Neight);
							}
						}
					}
				}
			}
		}

		for (int i=0; i<Sprites.Count; ++i) {
			TileScript Current = Sprites[i];

			GraphConnect GC = Current.Connections;
			
			for(int c=1; c<7; ++c) {
				MoveDir Dir = (MoveDir)c;
				bool IsConnected = GC.IsConnected(Dir);
				bool HasConnectedNeight = false;

				TileScript Neight = GetNeight (Current.PosX, Current.PosY, Dir);
				if(Neight != null) {
					GraphConnect NGC = Neight.Connections;
					MoveDir InvDir = InverseDir(Dir);
					HasConnectedNeight = NGC.IsConnected(InvDir);
				}

				if(Current.RedWarn[(int)Dir] != null) {

					Current.RedWarn[(int)Dir].SetActive(Current.IsEnergy && IsConnected && !HasConnectedNeight);
				}
			}
		}
	}

	public TileScript AddSprite(int x, int y, GraphConnect Connections) {

		if(x < 0 || x >= CellNumX || x < 0 || y >= CellNumY) {
			return null;
		}

		if (Cells[x][y] != null) {
			return null;
		}

		GameObject CurObj = new GameObject("Tile_"+x+"_"+y);
		SpriteRenderer CurSpriteRen = CurObj.AddComponent("SpriteRenderer") as SpriteRenderer;
		CurSpriteRen.sprite = MainSprite;

		CircleCollider2D CurCircle = CurObj.AddComponent("CircleCollider2D") as CircleCollider2D;
		CurCircle.radius = 0.44f;		

		TileScript CurScript = CurObj.AddComponent("TileScript") as TileScript;
		CurScript.MainScript = this;
		CurScript.TileObj = CurObj;
		CurScript.Connections = Connections;
		CurScript.PosX = x;
		CurScript.PosY = y;
		CurScript.transform.position = CalcSpritePosition(x,y);

		Sprites.Add(CurScript);

		Cells [x] [y] = CurScript;


		return CurScript;
	}

	Vector3 CalcSpritePosition(int x, int y) {

		float Decal = (y % 2) == 0 ? 0.0f : GridSize / 2.0f;

		Vector3 CurPos = new Vector3();
		CurPos.x = x * GridSize + Decal;
		CurPos.y = -y * GridSizeY;

		return CurPos;
	}

	bool AddBackSprite(int x, int y) {
		
		if(x < 0 || x >= CellNumX || y < 0 || y >= CellNumY) {
			return false;
		}

		GameObject CurObj = new GameObject("TileBack_"+x+"_"+y);
		SpriteRenderer CurSpriteRen = CurObj.AddComponent("SpriteRenderer") as SpriteRenderer;
		
		CurSpriteRen.sprite = MainSprite;
		CurSpriteRen.color = BackColor;

		Vector3 NewPos = CalcSpritePosition(x,y);
		NewPos.z = 1.0f;
		CurObj.transform.position = NewPos;
		
		return true;
	}

	public bool HasNeight(int x, int y, MoveDir Dir) {

		return GetNeight(x, y, Dir) != null;
	}

	public bool IsAmongNeight(int x, int y, TileScript CurTile) {
		if (CurTile == null) {
			return false;
		}

		for (int i=1; i<7; ++i) {
			MoveDir Dir = (MoveDir)i;
			if( GetNeight(x, y, Dir) == CurTile) {
				return true;
			}
		}

		return false;
	}

	public TileScript GetNeight(int x, int y, MoveDir Dir) {

		int NewX = x;
		int NewY = y;
		if (!GetMoveNewPos (x, y, Dir, ref NewX, ref NewY)) {
			return null;
		}

		return Cells [NewX] [NewY];
	}

	public TileScript GetNeight(int x, int y, MoveDir Dir, ref int OutX, ref int OutY) {

		if (!GetMoveNewPos (x, y, Dir, ref OutX, ref OutY)) {
			return null;
		}
		
		return Cells [OutX] [OutY];
	}

	public bool IsOutside(int x, int y) {

		if(x < 0 || x >= CellNumX || y < 0 || y >= CellNumY) {
			return true;
		}
		return false;
	}
	
	public bool GetMoveNewPos(int x, int y, MoveDir Dir, ref int OutX, ref int OutY) {
			
		int NewX = x;
		int NewY = y;
		
		if (Dir == MoveDir.Left) {
			NewX -= 1;
		} else if (Dir == MoveDir.Right) {
			NewX += 1;
		} else if (Dir == MoveDir.RDown) {
			NewX += (y%2) == 1 ? 1 : 0;
			NewY += 1;
		} else if (Dir == MoveDir.RTop) {
			NewX += (y%2) == 1 ? 1 : 0;
			NewY -= 1;
		} else if (Dir == MoveDir.LDown) {
			NewX -= (y%2) == 0 ? 1 : 0;
			NewY += 1;
		} else if (Dir == MoveDir.LTop) {
			NewX -= (y%2) == 0 ? 1 : 0;
			NewY -= 1;
		}
		
		if (NewX == x && NewY == y) {
			return false;
		}

		if (IsOutside (NewX, NewY)) {
			return false;
		}
		
		OutX = NewX;
		OutY = NewY;
		return true;
	}

	public bool GetMoveNewPos(TileScript Tile, MoveDir Dir, ref int OutX, ref int OutY) {
		
		return GetMoveNewPos (Tile.PosX, Tile.PosY, Dir, ref OutX, ref OutY);
	}

	public bool CanMoveDir(TileScript Tile, MoveDir Dir) {
		int NewX = 0;
		int NewY = 0;
		if (!GetMoveNewPos (Tile, Dir, ref NewX, ref NewY)) {
			return false;
		}

		if (Cells [NewX] [NewY] != null) {
			return false;
		}

		return true;
	}

	public bool MoveTile(TileScript Tile, MoveDir Dir, bool JustPrepare) {

		if (Tile.IsEnergy == false) {
			Tile.TriggerBlock();
			return false;
		}

		if (Tile.TryMotionDir != MoveDir.None) {
			if(Tile.TryMotionDir == Dir) {
				return true;
			} else {
				Tile.TriggerBlock();
				return false;
			}
		}

		Tile.TryMotionDir = Dir;

		int NewX = Tile.PosX;
		int NewY = Tile.PosY;

		if (!GetMoveNewPos (Tile, Dir, ref NewX, ref NewY)) {
			Tile.TryMotionDir = MoveDir.None;
			Tile.TriggerBlock();
			return false;
		}

		GraphConnect CGC = Tile.Connections;

		TileScript OtherTile = Cells [NewX] [NewY];
		if (OtherTile != null) {
			GraphConnect OGC = OtherTile.Connections;
			if(!CGC.IsConnected(Dir) || !OGC.IsConnected(InverseDir(Dir)) || !MoveTile(OtherTile, Dir, JustPrepare)) {
				Tile.TriggerBlock();
				OtherTile.TriggerBlock();
				Tile.TryMotionDir = MoveDir.None;
				return false;
			}
		}

		MoveDir InvDir = InverseDir (Dir);
		if (CGC.IsConnected (InvDir)) {
			TileScript ConnectedTile = GetNeight(Tile.PosX, Tile.PosY, InvDir);
			if(ConnectedTile != null) {
				GraphConnect OGC = ConnectedTile.Connections;
				if(OGC.IsConnected(Dir) && !MoveTile(ConnectedTile, Dir, JustPrepare)) {
					Tile.TriggerBlock();
					ConnectedTile.TriggerBlock();
					Tile.TryMotionDir = MoveDir.None;
					return false;
				}
			}
		}

		if (JustPrepare) {
			SpritesToMove.Add(Tile);
		} else {
			if (Cells [Tile.PosX] [Tile.PosY] == Tile) {
				Cells [Tile.PosX] [Tile.PosY] = null;
			}
			Cells [NewX] [NewY] = Tile;
			Tile.PosX = NewX;
			Tile.PosY = NewY;
			Tile.transform.position = CalcSpritePosition(NewX,NewY);

			NeedUpdateEnergy = true;
		}

		Tile.TryMotionDir = MoveDir.None;

		return true;
	}

	public bool RequestMoveTile(TileScript Tile, Vector2 DragDir, bool JustPrepare) {
		
		MoveDir MotionDir = MoveDir.Left;
		
		if (Mathf.Abs (DragDir.y) > 0.5f) {
			if(DragDir.x > 0.0f) {
				if(DragDir.y > 0.0f) {
					MotionDir = MoveDir.RTop;
				} else {
					MotionDir = MoveDir.RDown;
				}
			} else {
				if(DragDir.y > 0.0f) {
					MotionDir = MoveDir.LTop;
				} else {
					MotionDir = MoveDir.LDown;
				}
			}
		} else {
			if(DragDir.x > 0.0f) {
				MotionDir = MoveDir.Right;
			} else {
				MotionDir = MoveDir.Left;
			}
		}

		return MoveTile (Tile, MotionDir, JustPrepare);
	}

	void DragBehavior() {
		
		Vector3 Mouse3DPos = Camera.main.camera.ScreenToWorldPoint(Input.mousePosition);
		Mouse3DPos.z = 0.0f;
		Vector2 MousePos = new Vector2 (Mouse3DPos.x, Mouse3DPos.y);
		
		bool NearestCellFound = false;
		int NearestCellX = 0;
		int NearestCellY = 0;
		float NearestCellDist = 64.0f;
		
		for (int x = 0; x<Cells.Count; ++x) {
			for (int y = 0; y<Cells[x].Count; ++y) {
				
				Vector3 CellPos = CalcSpritePosition(x,y);
				CellPos.z = 0.0f;
				
				float CellDist = Vector3.Distance(Mouse3DPos, CellPos);
				if(CellDist < NearestCellDist) {
					NearestCellFound = true;
					NearestCellX = x;
					NearestCellY = y;
					NearestCellDist = CellDist;
					LastNearestPos = new Vector2(CellPos.x, CellPos.y);
				}
			}
		}
		
		TileScript NearestTile = null;
		
		if (NearestCellFound) {
			NearestTile = Cells [NearestCellX] [NearestCellY];
		}
		
		if(PrevOverideTile != null && NearestTile != PrevOverideTile) {
			PrevOverideTile.IsMouseOver = false;
		}
		
		if (NearestTile != null && NearestTile.IsEnergy) {
			NearestTile.IsMouseOver = true;
			
		}
		
		bool MouseClic = Input.GetMouseButton (0);
		if (MouseClic && !WasMouseClic) {
			DragTile = NearestTile;
			DragLastDir.Set (0.0f, 0.0f);
		}
		
		if (!MouseClic) {
			DragTile = null;
		}
		
		if(SpritesToMove.Count != 0) {
			for (int i=0; i<SpritesToMove.Count; ++i) {
				TileScript CurTile = SpritesToMove [i];
				if(CurTile != null) {
					CurTile.transform.position = CalcSpritePosition (CurTile.PosX, CurTile.PosY);
				}
			}
			SpritesToMove.Clear ();
		}
		
		if (DragTile != null) {
			
			Vector2 DragDir = MousePos - LastNearestPos;
			Vector2 MoveDirNorm = DragDir.normalized;
			
			Vector2 DragSnapDir = new Vector2 ();
			if (Mathf.Abs (MoveDirNorm.y) > 0.5f) {
				float Proj = Vector2.Dot (DragDir, TwistDir);
				DragSnapDir.x = Proj * TwistDir.x * Mathf.Sign(DragDir.x) * Mathf.Sign(DragDir.y);
				DragSnapDir.y = Proj * TwistDir.y;
			} else {
				DragSnapDir.x = DragDir.x;
			}
			
			float DragLength = DragDir.magnitude;
			if(DragLength > 0.1) {
				bool CanMove = RequestMoveTile (DragTile, MoveDirNorm, true);
				
				if(CanMove) {
					Vector3 NewPos = new Vector3 ();
					for (int i=0; i<SpritesToMove.Count; ++i) {
						TileScript CurTile = SpritesToMove [i];
						NewPos = CalcSpritePosition (CurTile.PosX, CurTile.PosY);
						
						NewPos.x += DragSnapDir.x;
						NewPos.y += DragSnapDir.y;
						
						CurTile.transform.position = NewPos;
					}
				}
			}
			
			if (NearestCellFound && DragTile != NearestTile) {
				
				Vector3 PrevPos = CalcSpritePosition (DragTile.PosX, DragTile.PosY);
				Vector3 NearPos = CalcSpritePosition (NearestCellX, NearestCellY);
				
				Vector2 NewMoveDir = new Vector2(NearPos.x - PrevPos.x, NearPos.y - PrevPos.y);
				NewMoveDir.Normalize();
				
				//if (IsAmongNeight (NearestCellX, NearestCellY, DragTile) && DragLastDir.sqrMagnitude > 0.1) {
				//if (IsAmongNeight (NearestCellX, NearestCellY, DragTile)) {
				if(true) {

					SaveState TmpState = new SaveState(Cells);

					bool WillMove = RequestMoveTile (DragTile, NewMoveDir, false);

					if(WillMove) {
						History.Add(TmpState);
						if(History.Count > 30) {
							History.RemoveAt(0);
						}
					}
					
					if(!WillMove) {
						DragTile = null;
						DragSnapDir.x = 0.0f;
						DragSnapDir.y = 0.0f;
					}
					
					Vector3 NewPos = new Vector3 ();
					for (int i=0; i<SpritesToMove.Count; ++i) {
						TileScript CurTile = SpritesToMove [i];
						NewPos = CalcSpritePosition (CurTile.PosX, CurTile.PosY);
						
						NewPos.x += DragSnapDir.x;
						NewPos.y += DragSnapDir.y;
						
						CurTile.transform.position = NewPos;
					}
				}/* else {
					DragTile = null;
				}*/
			}
			
			DragLastDir.Set ( MoveDirNorm.x, MoveDirNorm.y );
		}
		
		WasMouseClic = MouseClic;
		
		PrevOverideTile = NearestTile;
	}

	void TestVictoryConditions() {

		bool CanWin = true;

		for (int i = 0; i<Sprites.Count; ++i) {
			if(!Sprites[i].IsEnergy) {
				CanWin = false;
				break;
			}
		}

		if (CanWin) {
			for (int i = 0; i<Sprites.Count; ++i) {
				TileScript CurTile = Sprites[i];
				for (int d = 1; d<7; ++d) {
					MoveDir Dir = (MoveDir)d;
					TileScript Neight = GetNeight(CurTile.PosX, CurTile.PosY, Dir);
					if(Neight == null) {
						if(CurTile.Connections.G[(int)Dir]) {
							CanWin = false;
							break;
						}
					} else {
						MoveDir InvDir = InverseDir(Dir);
						bool CurConnect = CurTile.Connections.G[(int)Dir];
						bool NeightConnect = Neight.Connections.G[(int)InvDir];
						if(CurConnect != NeightConnect) {
							CanWin = false;
							break;
						}
					}
				}
				if(!CanWin) {
					break;
				}
			}
		}

		if (CanWin) {
			if(!HaveWon) {
				SoundPuzzleWin.audio.Play ();
			}
			HaveWon = true;
		}
	}

	void UpdateTileBasicPos() {
		for (int i=0; i<Sprites.Count; ++i) {
			TileScript CurTile = Sprites [i];
			CurTile.transform.position = CalcSpritePosition (CurTile.PosX, CurTile.PosY);
		}
	}

	// Update is called once per frame
	void FixedUpdate () {

		if (!IsInit || IsInHelp) {
			return;
		}

		TimeBlackScreen -= Time.fixedDeltaTime * 0.5f;
		if (TimeBlackScreen < 0.0f) {
			TimeBlackScreen = 0.0f;
		}

		if (!HaveWon) {
			DragBehavior ();

			TimeBlackScreen = 1.0f;

			BlackQuad.gameObject.SetActive(false);
		} else {

			if (PrevOverideTile != null) {
				PrevOverideTile.IsMouseOver = false;
				PrevOverideTile = null;
			}

			UpdateTileBasicPos();

			BlackQuad.gameObject.SetActive(true);
			SpriteRenderer BlackSprite = BlackQuad.renderer as SpriteRenderer;
			Color CurCol = BlackSprite.color;
			CurCol.a = 1.0f-TimeBlackScreen;
			BlackSprite.color = CurCol;
		}

		if (Time.fixedTime > ResetTime + 0.1f) {
			ResetTime = Time.fixedTime;
			++ResetNum;
			//StartCreateLevel();
			//ScrambleTest(30);
		}

		if(NeedUpdateEnergy) {
			UpdateEnergySprites ();
		}

		for (int i = 0; i<Sprites.Count; ++i) {
			Sprites[i].Tick();
		}

		TestVictoryConditions ();
	
		/*
		for (int x = 0; x<Cells.Count; ++x) {
			for (int y = 0; y<Cells[x].Count; ++y) {

				if(Cells[x][y] != null) {
					Cells[x][y].Tick();
				}
			}
		}*/

	}
}
